# `delay_samples` param is not reliably optimizable. See `delay_line.py` for more details.

import numpy as np
import jax.numpy as jnp
from jax import jit

NAME = 'Feedforward Delay'

MAX_DELAY_SIZE_SAMPLES = 44_100

def init_params(wet_amount=1.0, delay_samples=6.5):
    return {
        'wet_amount': wet_amount,
        'delay_samples': delay_samples,
    }

def init_state():
    return {}

def default_target_params():
    return init_params(0.5, 6.0)

def tick(carry, x):
    raise 'single-sample tick method not implemented for feedforward delay'

@jit
def tick_buffer(carry, X):
    params = carry['params']
    state = carry['state']
    delay_samples = params['delay_samples']
    remainder = delay_samples - jnp.floor(delay_samples)
    X_linear_interp = (1 - remainder) * X + remainder * jnp.concatenate([jnp.array([0]), X[:X.size - 1]])
    Y = jnp.where(jnp.arange(X.size) >= delay_samples, jnp.roll(X_linear_interp, delay_samples.astype('int32')), 0)
    return carry, X * (1 - params['wet_amount']) + Y * params['wet_amount']



import jax.numpy as jnp

def mse(X, Y):
    return ((Y - X) ** 2).mean()

def mae(X, Y):
    return jnp.abs(Y - X).mean()

# Based on https://github.com/numpy/numpy/issues/2310#issuecomment-314265666
def correlation(X, Y):
    c = jnp.correlate((X - jnp.mean(X)) / (jnp.std(X) * len(X)), (Y - jnp.mean(Y)) / (jnp.std(Y)))[0]
    return 1 - (1 + c) / 2

# TODO error-to-signal ratio (ESR)
#   MSE normalized by signal energy: "You need to be more accurate when your signal is very quiet"



import numpy as np
import jax.numpy as jnp
from jax import grad, value_and_grad, jit, vmap
from jax.experimental import optimizers
import collections.abc

from jaxdsp.processors import serial_processors
from jaxdsp.loss import mse, correlation
from jax.tree_util import tree_map, tree_multimap

@jit
def mean_loss_and_grads(loss, grads):
    return np.mean(loss), tree_map(lambda grad: np.mean(grad, axis=0), grads)

def process(processor, params, X, *init_state_args):
    carry, Y = processor.tick_buffer({'params': params, 'state': processor.init_state(*init_state_args)}, X)
    return Y

def evaluate(params_estimated, params_target, processor, X, *init_state_args):
    Y_estimated = process(processor, params_estimated, X, *init_state_args)
    Y_target = process(processor, params_target, X, *init_state_args)
    return Y_estimated, Y_target

def train(processors, Xs, step_size=0.2, num_batches=200, batch_size=32,
          params_init=None, params_target=None):
    processor = serial_processors
    params_target = params_target or processor.default_target_params(processors)
    def loss(params, X):
        Y_estimated, Y_target = evaluate(params, params_target, processor, X, processors)
        return mse(Y_estimated, Y_target)

    params_init = params_init or processor.init_params(processors)
    params_history = tree_map(lambda param: [param], params_init)
    loss_history = np.zeros(num_batches)
    grad_fn = jit(vmap(value_and_grad(loss), in_axes=(None, 0), out_axes=0))
    opt_init, opt_update, get_params = optimizers.sgd(step_size)
    opt_state = opt_init(params_init)
    for batch_i in range(num_batches):
        Xs_batch = Xs[np.random.choice(Xs.shape[0], size=batch_size)]
        loss, grads = mean_loss_and_grads(*grad_fn(get_params(opt_state), Xs_batch))
        # TODO clip grads such that all params are in [0, 1]
        opt_state = opt_update(batch_i, grads, opt_state)
        loss_history[batch_i] = loss
        new_params = get_params(opt_state)
        # params_history = tree_multimap(lambda h, *rest: h.append(rest[0]), params_history, new_params)
        for (processor_key, processor_params) in new_params.items():
            for (param_key, param) in processor_params.items():
                params_history[processor_key][param_key].append(param)

    params_estimated = get_params(opt_state)
    return params_estimated, params_target, params_history, loss_history



from collections.abc import Iterable
from matplotlib import pyplot as plt
import numpy as np

from jaxdsp.training import train


def plot_filter(X, Y, Y_reference, Y_estimated, title):
    column_titles = ['Target', 'Reference Implementation', 'Estimated'] if Y_reference is not None else ['Target', 'Estimated']
    row_titles = ['Input', 'Output']
    Ys = [Y, Y_reference, Y_estimated] if Y_reference is not None else [Y, Y_estimated]
    _, axes = plt.subplots(2, len(column_titles), figsize=(14, 6))
    plt.suptitle(title, size=16)
    for ax, column_title in zip(axes[0], column_titles):
        ax.set_title(column_title)
    for ax, row_title in zip(axes[:,0], row_titles):
        ax.set_ylabel(row_title, size='large')
    for i, ax_column in enumerate(axes.T):
        in_plot = ax_column[0]
        # in_plot.stem(X, basefmt=' ')
        in_plot.plot(X)
        in_plot.set_ylim([X.min() - 0.1, X.max() + 0.1])

        out_plot = ax_column[1]
        # out_plot.stem(Ys[i], basefmt=' ')
        out_plot.plot(Ys[i])
        out_plot.set_ylim([Ys[i].min() - 0.1, Ys[i].max() + 0.1])
    plt.tight_layout()

def plot_loss(loss_history):
    _, (plot) = plt.subplots(1, 1, figsize=(14, 3))
    plot.plot(loss_history)
    plot.set_xlabel('Batch')
    plot.set_ylabel('Loss')
    plot.set_title('Loss over time', size=16)
    plot.set_yscale('log', base=10)
    plot.autoscale(tight=True)

def plot_params_single(processor_name, params_target, params_history):
    param_groups = [] # list of lists of (key, label, scalar_param_value)
    # All non-list params go into the first (and potentially only) param group (column).
    single_params = [(key, key, param) for (key, param) in params_target.items() if not isinstance(param, Iterable)]
    if len(single_params) > 0: param_groups.append(single_params)
    # Each list param gets its own param group (column).
    for (key, params) in [(key, params) for (key, params) in params_target.items() if isinstance(params, Iterable)]:
        param_groups.append([(key, '${}_{}$'.format(key, i), param) for i, param in enumerate(params)])
    num_rows, num_cols = max([len(param_group) for param_group in param_groups]), len(param_groups)
    _, axes = plt.subplots(num_rows, num_cols, figsize=(14, num_rows * 2))
    if num_rows == 1 and num_cols == 1: axes = np.expand_dims(axes, axis=0)
    if len(axes.shape) == 1: axes = np.expand_dims(axes, axis=1)
    plt.suptitle('Estimated parameters for {}'.format(processor_name), size=16)
    for param_group_i, param_group in enumerate(param_groups):
        for param_i, (key, label, param) in enumerate(param_group):
            plot = axes[param_i][param_group_i]
            plot.set_title(label)
            plot.axhline(y=param, c='g', linestyle='--', label='Actual params')
            param_history = params_history[key]
            if isinstance(param_history[0], Iterable): param_history = np.asarray(param_history)[:,param_i]
            plot.plot(param_history, c='b', label='Estimated params')
            if param_group_i == 0: plot.set_ylabel('Value')
            if param_i == 0: plot.legend()
            if param_i == len(param_group) - 1: plot.set_xlabel('Batch')
            plot.autoscale()
    plt.tight_layout()

def plot_params(params_target, params_history):
    for processor_name in params_target.keys():
        plot_params_single(processor_name, params_target[processor_name], params_history[processor_name])

# Used to investigate the shape of the gradient curve for a single target value across
# a range of different initial values.
# `params_inits` should be a list of dicts - one for each parameter tree to initialize with
# `params_target` should be a single target params dict
def plot_optimization(processor, Xs, params_inits, params_target, steps=50):
    params_estimated = {key: [] for key in params_target.keys()}
    initial_losses = []
    losses = []
    x = [param_init['delay_samples'] for param_init in params_inits] # TODO only pass in one to vary
    for params_init in params_inits:
        params_estimated_i, _, _, loss_history = train([processor], Xs,
                                                       params_init={processor.NAME: params_init},
                                                       params_target={processor.NAME: params_target},
                                                       num_batches=steps, batch_size=4)
        for key, param_estimated in params_estimated_i[processor.NAME].items():
            params_estimated[key].append(param_estimated)
        initial_losses.append(loss_history[0])
        losses.append(loss_history[-1])
    _, axes = plt.subplots(len(params_target) + 2, 1, figsize=(14, 12))
    for param_i, (label, param_target) in enumerate(params_target.items()):
        estimated_params_plot = axes[param_i]
        estimated_params_plot.plot(x, params_estimated[label], linewidth=3)
        estimated_params_plot.set_title('Estimated {}'.format(label), size=18)
        estimated_params_plot.set_ylabel('Estimated after {} steps'.format(steps), size=11)
        estimated_params_plot.axhline(param_target, linestyle='--', c='r', label='Target index')
        estimated_params_plot.legend()
    initial_loss_plot = axes[-2]
    initial_loss_plot.plot(x, initial_losses, linewidth=3)
    initial_loss_plot.set_title('Initial loss', size=18)
    initial_loss_plot.set_ylabel('Initial MSE loss', size=11)
    optimized_loss_plot = axes[-1]
    optimized_loss_plot.plot(x, losses, linewidth=3)
    optimized_loss_plot.set_title('Optimized loss', size=18)
    optimized_loss_plot.set_xlabel('Initial guess for {} value'.format(label), size=16)
    optimized_loss_plot.set_ylabel('MSE loss after {} steps'.format(steps), size=11)
    for axis in axes:
        axis.grid(True)
    plt.tight_layout()


import jax.numpy as jnp
from jax import jit

NAME = 'Clip'

def init_state():
    return {}

def init_params(min_value=-0.9, max_value=0.9):
    return {
        'min': min_value,
        'max': max_value,
    }

def default_target_params():
    return init_params(-0.5, 0.5)

@jit
def tick(carry, x):
    params = carry['params']
    return carry, jnp.clip(x, params['min'], params['max'])

@jit
def tick_buffer(carry, X):
    return tick(carry, X)



# Fractional delay line using linear interpolation

# The `delay_samples` param cannot be optimized outside of the +/- 1 sample range between
# init and target.
# This is because the param maps directly to an `index_update` operation,
# which requires converting the parameter to an integer, which is a non-differentiable operation.
# Linear interpolation helps by allowing successful optimization within the +/- 1 sample
# range, but I have not found a solution that allows the gradient signal to propagate
# across the entire indexing range.
# (See "Differentiable array indexing" notebook for more details.)


import jax.numpy as jnp
import numpy as np
from jax import jit, lax
from jax.ops import index, index_update

NAME = 'Delay Line'

MAX_DELAY_LENGTH_SAMPLES = 44_100

def init_params(wet_amount=1.0, delay_samples=9.2):
    return {
        'wet_amount': wet_amount,
        'delay_samples': delay_samples,
    }

def init_state():
    return {
        'delay_line': jnp.zeros(MAX_DELAY_LENGTH_SAMPLES),
        'read_sample': 0.0,
        'write_sample': 0.0,
    }

def default_target_params():
    return init_params(0.5, 10.0)

@jit
def tick(carry, x):
    params = carry['params']
    state = carry['state']

    write_sample = state['write_sample']
    read_sample = state['read_sample']

    state['delay_line'] = index_update(state['delay_line'], index[write_sample].astype('int32'), x)

    read_sample_floor = read_sample.astype('int32')
    interp = read_sample - read_sample_floor
    y = (1 - interp) * state['delay_line'][read_sample_floor]
    y += (interp) * state['delay_line'][(read_sample_floor + 1) % MAX_DELAY_LENGTH_SAMPLES]

    state['write_sample'] += 1
    state['write_sample'] %= MAX_DELAY_LENGTH_SAMPLES
    state['read_sample'] += 1
    state['read_sample'] %= MAX_DELAY_LENGTH_SAMPLES

    out = x * (1 - params['wet_amount']) + y * params['wet_amount']
    return carry, out

@jit
def tick_buffer(carry, X):
    state = carry['state']
    params = carry['params']
    state['read_sample'] = (state['write_sample'] - jnp.clip(params['delay_samples'], 0, MAX_DELAY_LENGTH_SAMPLES)) % MAX_DELAY_LENGTH_SAMPLES
    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit, lax

from jaxdsp.processors import lowpass_feedback_comb_filter as comb
from jaxdsp.processors import allpass_filter as allpass

NAME = 'Freeverb'

fixed_gain = 0.015
scale_wet = 3.0
scale_dry = 2.0
scale_damp = 0.4
scale_room = 0.28
offset_room = 0.7

comb_tunings_l = [1116, 1188, 1277, 1356, 1422, 1491, 1557, 1617]
allpass_tunings_l = [556, 441, 341, 225]
stereo_spread = 23

# comb_tunings_l = [1, 2, 3, 4, 5, 6, 7, 9]
# allpass_tunings_l = [2, 3, 4, 5]

def create_comb_carry(buffer_size, feedback=0.5, damp=0.0):
    return {
        'state': comb.init_state(buffer_size=buffer_size),
        'params': comb.init_params(feedback=feedback, damp=damp),
    }

def create_allpass_carry(buffer_size, feedback=0.5):
    return {
        'state': allpass.init_state(buffer_size=buffer_size),
        'params': allpass.init_params(feedback=feedback),
    }

def init_state():
    return {
        'combs_l': [create_comb_carry(buffer_size) for buffer_size in comb_tunings_l],
        'combs_r': [create_comb_carry(buffer_size + stereo_spread) for buffer_size in comb_tunings_l],
        'allpasses_l': [create_allpass_carry(buffer_size) for buffer_size in allpass_tunings_l],
        'allpasses_r': [create_allpass_carry(buffer_size + stereo_spread) for buffer_size in allpass_tunings_l],
    }

def init_params():
    return {
        'wet': 0.0,
        'dry': 1.0,
        'width': 0.0,
        'damp': 0.0,
        'room_size': 0.0,
    }

def default_target_params():
    return {
        'wet': 0.3,
        'dry': 0.0,
        'width': 1.0,
        'damp': 0.5,
        'room_size': 0.5,
    }

@jit
def tick(carry, x):
    x_r, x_l = jnp.broadcast_to(x, (2,)) # handle stereo or mono in
    x_combined = (x_l + x_r) * fixed_gain

    state = carry['state']
    params = carry['params']

    out_l = 0.0
    out_r = 0.0

    # comb filters in parallel
    for i in range(len(comb_tunings_l)):
        state['combs_l'][i], out_comb_l = comb.tick(state['combs_l'][i], x_combined)
        state['combs_r'][i], out_comb_r = comb.tick(state['combs_r'][i], x_combined)
        out_l += out_comb_l
        out_r += out_comb_r
    # allpasses in series
    for i in range(len(allpass_tunings_l)):
        state['allpasses_l'][i], out_l = allpass.tick(state['allpasses_l'][i], out_l)
        state['allpasses_r'][i], out_r = allpass.tick(state['allpasses_r'][i], out_r)

    wet = params['wet'] * scale_wet
    dry = params['dry'] * scale_dry
    wet_1 = wet * (params['width'] / 2 + 0.5)
    wet_2 = wet * ((1 - params['width']) / 2)

    output_l = out_l * wet_1 + out_r * wet_2 + x_l * dry
    output_r = out_r * wet_1 + out_l * wet_2 + x_r * dry
    return carry, jnp.array([output_l, output_r])

@jit
def tick_buffer(carry, X):
    state = carry['state']
    params = carry['params']

    room_size = (params['room_size'] * scale_room) + offset_room
    damp = params['damp'] * scale_damp
    for comb_l in state['combs_l']:
        comb_l['params']['feedback'] = room_size
        comb_l['params']['damp'] = damp
    for comb_r in state['combs_r']:
        comb_r['params']['feedback'] = room_size
        comb_r['params']['damp'] = damp

    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit, lax

NAME = 'FIR Filter'

def init_state(length=4):
    return {
        'inputs': jnp.zeros(length)
    }

def init_params(length=4):
    return {
        'B' : jnp.concatenate([jnp.array([1.0]), jnp.zeros(length - 1)])
    }

def default_target_params(length=4):
    return {
        'B': jnp.array([0.1, 0.7, 0.5, 0.6])
    }

@jit
def tick(carry, x):
    state = carry['state']
    params = carry['params']
    B = params['B']
    state['inputs'] = jnp.concatenate([jnp.array([x]), state['inputs'][0:-1]])
    y = B @ state['inputs']
    return carry, y

@jit
def tick_buffer(carry, X):
    params = carry['params']
    B = params['B']
    return carry, jnp.convolve(X, B)[:-(B.size - 1)]
    # Impossibly, the following seems to perform about the exact same or even faster for large N?! O_O
    # return lax.scan(tick, carry, X)[1]



NAME = 'Serial Processors'

def init_state(processors):
    return {processor.NAME: { 'state': processor.init_state(), 'tick_buffer': processor.tick_buffer } for processor in processors}

def init_params(processors):
    return {processor.NAME: processor.init_params() for processor in processors}

def default_target_params(processors):
    return {processor.NAME: processor.default_target_params() for processor in processors}

def tick_buffer(carry, X):
    state = carry['state']
    params = carry['params']
    Y = X
    # XXX this quite right. need to clean up `carry` API
    for processor_name in state.keys():
        processor_carry = {'state': state[processor_name]['state'], 'params': params[processor_name]}
        processor_carry, Y = state[processor_name]['tick_buffer'](processor_carry, Y)
        # state[processor_name] = processor_carry['state']
        # params[processor_name] = processor_carry['params']
    return carry, Y



import jax.numpy as jnp
from jax import jit, lax
from jax.ops import index_update

NAME = 'Lowpass Feedback Comb Filter'

def init_params(feedback=0.0, damp=0.0):
    return {
        'feedback': feedback,
        'damp': damp,
    }

def init_state(buffer_size=20):
    return {
        'buffer': jnp.zeros(buffer_size),
        'buffer_index': 0,
        'filter_store': 0.0,
    }

def default_target_params():
    return init_params(0.5, 0.5)

@jit
def tick(carry, x):
    params = carry['params']
    state = carry['state']

    out = state['buffer'][state['buffer_index']]
    state['filter_store'] = out * (1 - params['damp']) + state['filter_store'] * params['damp']

    state['buffer'] = index_update(state['buffer'], state['buffer_index'], x + state['filter_store'] * params['feedback'])
    state['buffer_index'] += 1
    state['buffer_index'] %= state['buffer'].size
    return carry, out

@jit
def tick_buffer(carry, X):
    return lax.scan(tick, carry, X)



import jax.numpy as jnp
from jax import jit, lax
from scipy import signal

NAME = 'IIR Filter'

def init_state(length=5):
    return {
        'inputs': jnp.zeros(length),
        'outputs': jnp.zeros(length - 1),
    }

def init_params(length=5):
    return {
        'B' : jnp.concatenate([jnp.array([1.0]), jnp.zeros(length - 1)]),
        'A' : jnp.concatenate([jnp.array([1.0]), jnp.zeros(length - 1)]),
    }

def default_target_params(length=5):
    B, A = signal.butter(length - 1, 0.5, 'low')
    return {
        'B': B,
        'A': A,
    }

@jit
def tick(carry, x):
    params = carry['params']
    state = carry['state']
    B = params['B']
    A = params['A']
    state['inputs'] = jnp.concatenate([jnp.array([x]), state['inputs'][0:-1]])
    y = B @ state['inputs']
    if state['outputs'].size > 0:
        y -= A[1:] @ state['outputs']
        # Don't optimize the output gain, since it's not commonly used and constraining it to 1 helps training
        # Note that this makes the implementation technically incorrect. We can always uncomment if we want it back.
        # y /= A[0]
        state['outputs'] = jnp.concatenate([jnp.array([y]), state['outputs'][0:-1]])
    return carry, y

@jit
def tick_buffer(carry, X):
    return lax.scan(tick, carry, X)

